@@ -1,3 +1,4 @@
+# $Id: test.pl,v 1.2 2003/10/10 20:53:15 jacob Exp $
 # Before `make install' is performed this script should be runnable with
 # `make test'. After `make install' it should work as `perl test.pl'
 
@@ -8,7 +9,7 @@
 
 BEGIN { $| = 1; print "1..1\n"; }
 END {print "not ok 1\n" unless $loaded;}
-use Apache::AuthCookieDBI;
+# use Apache::AuthCookieDBI; # commented; can't really run this without Apache
 $loaded = 1;
 print "ok 1\n";
 
